/*
 *  GTRunLoopDelegate.h
 *  All
 *
 *  Created by nhn on 11. 1. 1..
 *  Copyright 2011 nhn KEH. All rights reserved.
 *
 */
#pragma once

#include <list>
#include <map>
#include <pthread.h>
#include "IGTEventDelegate.h"
#include "TypeDefinition.h"
#include "GTLog.h"

class CGTEventDelegate : public IGTEventDelegate
{
public:
	
	CGTEventDelegate(IGTEventDelegateEvent *pEvent);
	virtual ~CGTEventDelegate(void);
	
	virtual bool CanDelete(void);
	virtual void Delete(void);
	virtual void ReleaseAllSourcesAndTimers(void);

	///////////////////////////////////////////////////////
	// Observer
	///////////////////////////////////////////////////////
	virtual void RegisterObserver(IGTRunLoopObserver *pObserver);
	virtual void UnRegisterObserver(IGTRunLoopObserver *pObserver);
	
	///////////////////////////////////////////////////////
	// Timer 
	///////////////////////////////////////////////////////
	virtual GT_EVENT_HANDLE SetTimer(GT_EVENT_HANDLE handle, int msec, GT_PARAM param);
	virtual void KillTimer(GT_EVENT_HANDLE handle);
	
	///////////////////////////////////////////////////////
	// Custom Source
	///////////////////////////////////////////////////////
	virtual void RegisterSourceToCurrentRunLoop(int nPositionID, GT_EVENT_HANDLE handle);
	virtual void UnRegisterSourceToCurrentRunLoop(int nPositionID, GT_EVENT_HANDLE handle);	
	virtual void PostSourceEvent(GT_EVENT_HANDLE handle, GT_PARAM param1, GT_PARAM param2);
	virtual void PostSourceEventEx(int nPositionID, GT_EVENT_HANDLE handle, GT_PARAM param1, GT_PARAM param2);
		
	// this function does not allow to be called exception of 
	// <private>
public:	
	void _FireSourceEvent(GT_EVENT_HANDLE handle, GT_PARAM param1, GT_PARAM param2);
	void _FireTimerEvent(GT_EVENT_HANDLE handle, GT_PARAM param);
	void _Retain();
	void _Release(bool bRetainCountChange = true);
	
	// Thread Safety
private:
	void _PushToNewbieList(IGTRunLoopObserver *observer);
	void _PushToGarbageList(IGTRunLoopObserver *observer);
	bool _FindGarbageList(IGTRunLoopObserver *observer);
	void _ApplyObserverList();
	
	
private:
	typedef std::list<IGTRunLoopObserver *>  listobserver_t;
	typedef std::map<int, void *>	listprivate_t;
	typedef std::map<GT_EVENT_HANDLE, listprivate_t *> listsourcetarget_t;
	
	void *m_pDelegateImpl;
	pthread_mutex_t m_mutexStatus;			// lock order 1
	pthread_mutex_t m_mutexObserver;		// lock order 2
	pthread_mutex_t m_mutexObserverTemp;	// lock order 3
	pthread_mutex_t m_mutexSourceTarget;	// lock order 3

	listobserver_t m_listObservers;
	listobserver_t m_listObservers_Garbage;
	listobserver_t m_listObservers_Newbie;

	listsourcetarget_t m_mapSourceTarget;
	
	enum check_status { EXIST = 1, NOTHING = 0 };
	volatile int32_t m_nCheckReadyObserver;	
	volatile int32_t m_nSourceTargetCount;

	enum delegate_status { ACTIVE = 0, RELEASE = 1, DELETE = 2};
#define BEGIN_BLOCK_THREAD_SAFETY	\
	LOCK(m_mutexStatus);			\
	if(ACTIVE == m_nStatusMark)		\
	{						

#define END_BLOCK_THREAD_SAFETY		\
	}								\
	UNLOCK(m_mutexStatus);
	
	volatile enum delegate_status m_nStatusMark;
	volatile bool m_bObserverFireDoing;
	volatile int32_t m_nRetainCount;
	
	IGTEventDelegateEvent *m_pEvent;
};


inline bool CGTEventDelegate::CanDelete(void)
{
	if(0 == m_nRetainCount && ACTIVE != m_nStatusMark)
		return true;
	return false;
}


inline void CGTEventDelegate::_Retain()
{
	InterlockedIncrement(&m_nRetainCount);
	
	//GT_PLOG(">> CGTEventDelegate::_Retain() : %p Retain Count = %d, thread = %p", this, m_nRetainCount, pthread_self());
}


inline void CGTEventDelegate::_Release(bool bRetainCountChange)
{
	if(bRetainCountChange)
		InterlockedDecrement(&m_nRetainCount);	
	
#if 1	
	if(m_nRetainCount < 0)
	{
		GT_PLOG(">> CGTEventDelegate::_Release() : !!!!!!!!!!!!!!!!! ERROR !!!!!!!!!!!!! Retain Count = %d %p", m_nRetainCount, this);		
		ASSERT(false);
	}
#endif
	
	//GT_PLOG(">> CGTEventDelegate::_Release() : %p Retain Count = %d thread = %p", this, m_nRetainCount, pthread_self());
	if(0 == m_nRetainCount)
	{	
		switch(m_nStatusMark)
		{
			case DELETE:
			{
				GT_PLOG(">> CGTEventDelegate::_Release() : m_nStatusMark == DELETE!! %p thread %p", this, pthread_self());
				delete this;
				break;
			}
			default:
			{
				// TODO..
				break;
			}
		}
	}
	else if(1 == m_nRetainCount)
	{
		switch(m_nStatusMark)
		{
			case RELEASE:
			{
				GT_PLOG(">> CGTEventDelegate::_Release() : m_nStatusMark == RELEASE!! %p thread %p", this, pthread_self());

				if(m_pEvent)
					m_pEvent->OnGTEventDelegateEvent_DidReleased(this);
				break;
			}
			default:
			{
				// TODO
				break;
			}
		}
	}
}

